"""
This package contains serializers. Purpose of serializer class
is to convert hwt representations of designed architecture
to target language or form (VHDL/Verilog/SystemC...).
"""

# [TODO] move to hdlConvertor